#pragma once
#include <vector>
#include <string>
#include <chrono>
#include "ffgl/PluginInfo.h"
#include "ffgl/PluginBase.h"
#include "ffgl/ParamBase.h"
#include "ffgl/ParamFloat.h"
#include "ffgl/ParamText.h"
#include "ffgl/FFGLLib.h"
#include "ffglex/Shader.h"
#include "ffglex/ScreenQuad.h"
#include "ffglex/Utilities.h"
#include "ffglex/ScopedShaderBinding.h"
#include "ffglquickstart/ParamFFT.h"
#include "ParamsHSBAColor.h"
#include "ParamsRGBColor.h"
#include "ParamsVec.h"
#include "Utils.h"
#include "Audio.h"

namespace ffglqs
{

typedef ffgl::PluginInfo PluginInstance;

class Plugin : public ffgl::PluginBase
{
public:
	Plugin();
	~Plugin();

	/// This method allows to create the plugin description.
	/// This is what makes your plugin available in the exported library file.
	template< typename PluginType >
	static PluginInstance CreatePlugin( ffgl::PluginInfo infos, ffgl::PluginType type );

	/// Implementation of the FFGL InitGL instance specific function. This function allocates
	/// the OpenGL resources the plugin needs during its lifetime. If the shader fails to
	/// compile we return FF_FAIL.
	///
	/// \param	viewPort		Pointer to a ViewportStruct structure (see the definition in FFGL.h
	///							and the description in the FFGL specification).
	/// \return					This implementation returns FF_SUCCESS if the shader did compile
	///							correctly FF_FAIL otherwise.
	ffgl::FFResult InitGL( const ffgl::ViewportStruct* viewPort ) override;

	/// Implementation of the FFGL ProcessOpenGL instance specific function. It is in this function that
	/// the actual rendering occur.
	///
	/// \param	processOpenGLData	This is a parameter containing info about input texture if there is some.
	///
	/// \return		This implementation always returns FF_SUCCESS.
	ffgl::FFResult ProcessOpenGL( ffgl::ProcessOpenGLStruct* processOpenGLData ) override;

	/// Implementation of the FFGL DeinitGL instance specific function. This function frees
	/// any OpenGL resources the plugin has allocated
	///
	/// \return					This implementation always returns FF_SUCCESS.
	ffgl::FFResult DeinitGL() override;

	/// You can override this function to allocate any ressource you will need, shader for example. If any error
	/// occured (shader fails to compile for example), return FF_FAIL.
	/// \return                 FF_FAIL if any error occured else FF_SUCCESS
	virtual ffgl::FFResult Init()
	{
		return ffgl::SuccessFail::FF_SUCCESS;
	};

	/// You can override this function to do something just before the main shader call. It can for example be usefull
	/// if you have to react to certains parameter change in a specific way and / or send custom uniform to the shader
	virtual void Update(){};

	/// The default implementation will just call glDraw with the main shader. You can override this if your plugin needs
	/// to do more complex things, like calling multiple shaders.
	/// \return                 FF_FAIL if any error occured else FF_SUCCESS
	virtual ffgl::FFResult Render( ffgl::ProcessOpenGLStruct* inputTextures );

	/// Use this function to free any ressource allocated in your init() function
	virtual void Clean(){};

	/// This function allows to Include snippet of code that are provided in Utils.h. This allows plugin
	/// programmer to not rewrite common code in each plugin. For example, code to get a random number,
	/// map a value from a range to another, and simplex noise are currently available.
	/// \param	snippet		The id of a snippet to Include.
	void Include( shader::snippet_id snippet );

	/// This function will call the function just above for each snippets in the set.
	/// \param	snippets		A set of snippets to Include.
	void Include( std::set< shader::snippet_id > snippets );

	/// In this function we assemble the different bit of the fragment shader into one and try to compile it.
	/// This allows to add all the default uniform to the begining of the shader so you don't
	/// have to do it manually. It also permits to add any included snippet to it.
	/// \param	base		The base of your shader, usually just the void main() function
	/// \return				The full shader as a string
	std::string CreateFragmentShader( const std::string& base ) const;

	/// This function updates the uniform locations the plugin's parameters
	void Plugin::SetParamUniformLocations();

	/// This function is called by the plugin before each Render it allows to Update all the params that
	/// are sent to the main shader.
	void UpdateAudioAndTime();

	/// This function will send all the parameter registered by your plugin to the shader.
	/// \param	shader		The shader to send the params
	void SendParams( ffglex::Shader& shader ) const;

	/// This function will send all the default parameter (like time, bpm, resolution...) to the shader.
	/// \param	shader		The shader to send the params
	void SendDefaultParams( ffglex::Shader& shader ) const;

	/// This function is called by the host when the current sample rate has changed.
	/// \param	sampleRate	The new sample rate
	void SetSampleRate( unsigned int sampleRate ) override;

	/// This function must be called in the constructor of your plugin. It will allows to
	/// construct the full fragment shader. In order to simplify the development of plugin
	/// it is InitGL that takes care of defining all the parameters so that the shader can acces
	/// them.
	/// \param	base		The fragment shader
	void SetFragmentShader( const std::string& base );

	/// This function handle the special case where the parameter is a ParamFFT, and adds it to the audio vector as well as params
	/// \param	param		The parameter to add
	void AddAudioParam( ParamFFT* fft );

	/// This function allows to create a Hue color param, for exemple in Resolume this will display a
	/// color picker, which is very handy to choose your color.
	/// \param	name		The name of the parameter to add
	/// \param defaultColor	The default HSBA colour settinggs for the parameter
	/// \return				A ParamsVec struct filled in with the params created
	ParamsVec AddHSBAColorParams( const std::string& name, const ffglex::HSBAColor& defaultColor = ffglex::HSBAColor::RED );

	/// This function allows to create a Hue color param, for exemple in Resolume this will display a
	/// color picker, which is very handy to choose your color.
	/// \param	name		The base name of the parameters to add; internal names will be generated for saturation, brightness, alpha
	/// \param  uniformName The name of the vec3 uniform in the shader code
	/// \param defaultColor	The default values for the HSBA parameters
	/// \return				A ParamsVec struct filled in with the params created
	ParamsVec AddHSBAColorParams( const std::string& name, const std::string& uniformName, const ffglex::HSBAColor& defaultColor = ffglex::HSBAColor::RED );

	/// This function allows to create a Hue color param, for exemple in Resolume this will display a
	/// color picker, which is very handy to choose your color.
	/// \param	hsba		Reference to a ParamsVec struct to fill in with the params created
	/// \param	name		The name of the parameter to add
	/// \param  defaultColor  The default HSBA colour settinggs for the parameter
	void AddHSBAColorParams( ParamsVec& hsba, const std::string& name, const ffglex::HSBAColor& defaultColor = ffglex::HSBAColor::RED );

	/// This function allows to create a Hue color param, for exemple in Resolume this will display a
	/// color picker, which is very handy to choose your color.
	/// \param	hsba		Reference to a ParamsVec struct to fill in with the params created
	/// \param	name		The base name of the parameters to add; internal names will be generated for saturation, brightness, alpha
	/// \param  uniformName The name of the vec3 uniform in the shader code
	/// \param defaultColor	The default values for the HSBA parameters
	void AddHSBAColorParams( ParamsVec& hsba, const std::string& name, const std::string& uniformName, const ffglex::HSBAColor& defaultColor = ffglex::HSBAColor::RED );

	/// This function allows to create R,G,B color params. In Resolume it is displayed as
	/// separate sliders for each Red, Green and Blue.
	/// \param	name		The name of the parameter to add
	/// \param defaultColor	The default values for the HSBA parameters
	/// \return				A ParamsVec struct filled in with the params created
	ParamsVec AddRGBColorParams( const std::string& name, const ffglex::RGBAColor& defaultColor = ffglex::RGBAColor::GREY );

	/// This function allows to create R,G,B color params. In Resolume it is displayed as
	/// separate sliders for each Red, Green and Blue.
	/// \param	name		The base name for the three parameters; " R", " G", " B" will be appended
	/// \param  uniformName The name of the vec3 uniform in the shader code
	/// \param defaultColor	The default values for the HSBA parameters
	/// \return				A ParamsVec struct filled in with the params created
	ParamsVec AddRGBColorParams( const std::string& name, const std::string& uniformName, const ffglex::RGBAColor& defaultColor = ffglex::RGBAColor::GREY );

	/// This function allows to create R,G,B color params. In Resolume it is displayed as
	/// separate sliders for each Red, Green and Blue.
	/// \param	rgb			Reference to a ParamsVec struct to fill in with the params created
	/// \param	name		The name of the parameter to add
	/// \param defaultColor	The default values for the HSBA parameters
	void AddRGBColorParams( ParamsVec& rgb, const std::string& name, const ffglex::RGBAColor& defaultColor = ffglex::RGBAColor::GREY );

	/// This function allows to create R,G,B color params. In Resolume it is displayed as
	/// separate sliders for each Red, Green and Blue.
	/// \param	rgb			Reference to a ParamsVec struct to fill in with the params created
	/// \param	name		The base name for the three parameters; " R", " G", " B" will be appended
	/// \param  uniformName The name of the vec3 uniform in the shader code
	/// \param defaultColor	The default values for the HSBA parameters
	void AddRGBColorParams( ParamsVec& rgb, const std::string& name, const std::string& uniformName, const ffglex::RGBAColor& defaultColor = ffglex::RGBAColor::GREY );

	/// This function allows to create R,G,B,A color params. In Resolume they are displayed as
	/// separate sliders for each Red, Green, Blue and Alpha.
	/// \param	name		The base name for the four parameters; " R", " G", " B", " A" will be appended
	/// \param defaultColor	The default values for the HSBA parameters
	/// \return				A ParamsVec struct filled in with the params created
	ParamsVec AddRGBAColorParams( const std::string& name, const ffglex::RGBAColor& defaultColor = ffglex::RGBAColor::GREY );

	/// This function allows to create R,G,B,A color params. In Resolume they are displayed as
	/// separate sliders for each Red, Green, Blue and Alpha.
	/// \param	name		The base name for the four parameters; " R", " G", " B", " A" will be appended
	/// \param  uniformName The name of the vec4 uniform in the shader code
	/// \param defaultColor	The default values for the HSBA parameters
	/// \return				A ParamsVec struct filled in with the params created
	ParamsVec AddRGBAColorParams( const std::string& name, const std::string& uniformName, const ffglex::RGBAColor& defaultColor = ffglex::RGBAColor::GREY );

	/// This function allows to create R,G,B,A color params. In Resolume they are displayed as
	/// separate sliders for each Red, Green, Blue and Alpha.
	/// \param	rgba		Reference to a ParamsVec struct to fill in with the params created
	/// \param	name		The base name for the four parameters; " R", " G", " B", " A" will be appended
	/// \param defaultColor	The default values for the HSBA parameters
	void AddRGBAColorParams( ParamsVec& rgba, const std::string& name, const ffglex::RGBAColor& defaultColor = ffglex::RGBAColor::GREY );

	/// This function allows to create R,G,B,A color params. In Resolume they are displayed as
	/// separate sliders for each Red, Green, Blue and Alpha.
	/// \param	rgba		Reference to a ParamsVec struct to fill in with the params created
	/// \param	name		The base name for the four parameters; " R", " G", " B", " A" will be appended
	/// \param  uniformName The name of the vec4 uniform in the shader code
	/// \param defaultColor	The default values for the HSBA parameters
	void AddRGBAColorParams( ParamsVec& rgba, const std::string& name, const std::string& uniformName, const ffglex::RGBAColor& defaultColor = ffglex::RGBAColor::GREY );

	/// This function allows to create a RGB color param (actually three separate parameters). In Resolume each parameter is displayed as
	/// a colour gradient with values shown between 0 and 255 (with the middle value being 127.5).
	/// The actual values received from Resolume are in the standard [0,1], which is then mapped
	/// to the range specified by the RangeStruct range by GetFloatValue().
	/// \param	name		The base name for the three parameters; " R", " G", " B" will be appended
	/// \param  uniformName The name of the vec3 uniform in the shader code
	/// \param  defaultValue The default value for ALL of the RGB parameters
	/// \param  range       The range that the raw [0,1] values will be scaled to
	/// \return             Struct containing pointers to the three ParamRange parameters created (downcast to ParamFloat*)
	ParamsVec AddRGBRangeParams( const std::string& name, const std::string& uniformName, float defaultValue, const ffgl::RangeStruct range );

	/// This function allows to create a RGB color param (actually three separate parameters). In Resolume each parameter is displayed as
	/// a colour gradient with values shown between 0 and 255 (with the middle value being 127.5).
	/// The actual values received from Resolume are in the standard [0,1], which is then mapped
	/// to the range specified by the RangeStruct range by GetFloatValue().
	/// \param	rgb			Reference to a ParamsVec struct to fill in with the params created
	/// \param	name		The base name for the three parameters; " R", " G", " B" will be appended
	/// \param  uniformName The name of the vec3 uniform in the shader code
	/// \param  defaultValue The default value for ALL of the RGB parameters
	/// \param  range       The range that the raw [0,1] values will be scaled to
	/// \return             Struct containing pointers to the three ParamRange parameters created (downcast to ParamFloat*)
	void AddRGBRangeParams( ParamsVec& rgb, const std::string& name, const std::string& uniformName, float defaultValue, const ffgl::RangeStruct range );

	/// This function allows to check if a certains index corresponds to a Hue color. Internally, it is
	/// represented as three Params. It checks if the current one and the two followings are actually
	/// corresponding to a Hue color.
	/// \param	index		The index of the first parameter that could correspond to a Hue color.
	bool IsHueColor( int index ) const;

	/// This function allows to check if a certains index corresponds to a RGB color. Internally, it is
	/// represented as three Params. It checks if the current one and the two followings are actually
	/// corresponding to a RGB color.
	/// \param	index		The index of the first parameter that could correspond to a RGB color.
	bool IsRGBColor( int index ) const;

	/// This function allows to get a parameter by it's name.
	/// \param	name		The name of the parameter to get.
	/// \return				The parameter if we managed to find it.
	//const ParamBase* GetParam( std::string& name ) const;

	/// This function allows to get an option parameter by it's name.
	/// \param	name		The name of the parameter to get.
	/// \return				The parameter if we managed to find it.
	//ParamOption* getParamOption( const std::string& name );

	/// This function allows to get a text parameter by it's name.
	/// \param	name		The name of the parameter to get.
	/// \return				The parameter if we managed to find it.
	//ParamText* getParamText( const std::string& name );

protected:
	std::string fragmentShaderBase;
	std::vector< ParamFFT* > audioParams;
	ffglex::Shader shader;
	ffglex::ScreenQuad quad;

	float timeNow                                                         = 0;
	float deltaTime                                                       = 0;
	float lastUpdate                                                      = 0;
	int frame                                                             = 0;
	std::chrono::time_point< std::chrono::high_resolution_clock > t_start = std::chrono::high_resolution_clock::now();
	utils::Random random;
	std::set< shader::snippet_id > includedSnippets;

	std::string fragmentShaderCodeStart = R"(#version 410 core
out vec4 fragColor;
uniform vec2 resolution;
uniform float time;
uniform float deltaTime;
uniform int frame;
uniform float bpm;
uniform float phase;
)";
	std::string vertexShaderCode;

	// This function will consume the value for all ParamTrigger
	void ConsumeAllTriggers();
};

template< typename PluginType >
inline PluginInstance Plugin::CreatePlugin( ffgl::PluginInfo infos, ffgl::PluginType type )
{
	return PluginInstance(
		PluginFactory< PluginType >,// Create method, allows the host to create a new instance of the plugin
		infos.id.c_str(),           // Plugin unique ID
		infos.name.c_str(),         // Plugin name
		2,                          // API major version number
		1,                          // API minor version number
		infos.majorVersion,         // Plugin major version number
		infos.minorVersion,         // Plugin minor version number
		type,                       // Plugin type
		infos.description.c_str(),  // Plugin description
		infos.about.c_str()         // About
	);
}

}// namespace ffglqs
