// spxlSubAdd Effect Plugin
// More or less the same result as the AddSubtract sample plugin, with extra parameters for testing purposes.
// The "multiplier" parameter is used to scale the R,G,B addition/subtraction.

// 2019-08-12 ffgl@subpixels.com
// - Created, based on AddSubbtract sample code

#include "spxlSubAdd.h"
#include "ffgl/ParamBool.h"
#include "ffgl/ParamBuffer.h"
#include "ffgl/ParamFloat.h"
#include "ffgl/ParamInteger.h"
#include "ffgl/ParamOption.h"
#include "ffgl/ParamRange.h"
#include "ffgl/ParamText.h"
#include "ffglquickstart/ParamEvent.h"
#include "ffglquickstart/ParamFFT.h"
#include "ffglquickstart/ParamTrigger.h"

using namespace ffgl;
using namespace ffglex;
using namespace ffglqs;

namespace spxl
{
namespace ffgl
{
namespace effect
{
static ::ffgl::PluginInfo pluginInfo(
	PluginFactory< SubAdd >,// Create method
#ifdef _DEBUG
	"?e01",   // Plugin unique ID of maximum length 4.
	"?SubAdd",// Plugin name
#else
	"_e01",   // Plugin unique ID of maximum length 4.
	"_SubAdd",// Plugin name
#endif
	2,                    // API major version number
	1,                    // API minor version number
	1,                    // Plugin major version number
	0,                    // Plugin minor version number
	PluginType::FF_EFFECT,// Plugin type

	"Subtract and Add colours",                        // Plugin description
	"Resolume FFGL Example (mod by ffgl@subpixels.com)"// About
);

static const char _fragmentShaderCode[] = R"(

// All components are in the range [01], including hue.
vec3 rgb2hsv(vec3 c)
{
    vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
    vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
    vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

    float d = q.x - min(q.w, q.y);
    float e = 1.0e-10;
    return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
}

// All components are in the range [01], including hue.
vec3 hsv2rgb(vec3 c)
{
    vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
    vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
    return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

void main()
{
	vec4 rgba = texture( inputTexture, i_uv );
	vec4 orig = rgba;

	//The InputTexture contains premultiplied colors, so we need to unpremultiply first to apply our effect on straight colors.
	if( rgba.a > 0.0 )
		rgba.rgb /= rgba.a;

	vec3 hsb = rgb2hsv(rgba.rgb);

	rgba.rgb += adjustment * scaleAdj;

	if (doAdd) rgba.rgb += colAdd.rgb * scaleAdd;
	if (doSub) rgba.rgb -= colSub.rgb * scaleSub;

	float mask = 1.0f;

	if (hue.w > 0.0)
	{
		// Hue distance wraps ends (max distance is 0.5)
		float dist = abs(hue.x - hsb.x);
		dist = min(dist, 1.0 - dist);

		// Smoothstep up from below smooth and halfwidth below 1 to halfwidth below 1
		float match = smoothstep(1.0 - (hue.y / 2.0) - hue.z, 1.0 - hue.y / 2.0, 1.0 - dist);
		mask *= mix(1.0, match, hue.w);
	}

	if (sat.w > 0.0)
	{
		// Smoothstep (up) below min sat, smoothstep (down) above max sat
		float match = smoothstep(sat.x - sat.z, sat.x, hsb.y) * (1.0 - smoothstep(sat.y, sat.y + sat.z, hsb.y));
		mask *= mix(1.0, match, sat.w);
	}

	if (bri.w > 0.0)
	{
		// Smoothstep (up) below min bri, smoothstep (down) above max bri
		float match = smoothstep(bri.x - bri.z, bri.x, hsb.z) * (1.0 - smoothstep(bri.y, bri.y + bri.z, hsb.z));
		mask *= mix(1.0, match, bri.w);
	}

	// Show the mask, or use to mix the color
	if (showMask)
		rgba = vec4(mask, mask, mask, 1.0);
	else
		rgba.rgb = mix(orig.rgb, rgba.rgb, mask);

	// Can also etch using the 
	if (cutoutMask)
		rgba.a *= mask;

	//The plugin has to output premultiplied colors, this is how we're premultiplying our straight color while also
	//ensuring we aren't going out of the LDR the video engine is working in.
	rgba.rgb = clamp( rgba.rgb * rgba.a, vec3( 0.0 ), vec3( rgba.a ) );
	fragColor = rgba;
}
)";

SubAdd::SubAdd()
{
	// We declare that this plugin has a Brightness parameter which is a RGB param.
	// The uniform name ("adjustment") must match the uniform name in the fragment shader.
	AddRGBRangeParams(adjustment, "Adjust", "adjustment", 0.0f, { -1.0f, 1.0f } );

	AddParam( new ParamFloat( "Scale", "scaleAdj", 1.0f ) );

	colAdd = AddHSBAColorParams( "Add", "colAdd", HSBAColor::BLACK );
	colSub = AddHSBAColorParams( "Subtract", "colSub", HSBAColor::BLACK );

	AddParam( new ParamFloat( "Add Scale", "scaleAdd", 1.0f ) );
	AddParam( new ParamFloat( "Sub Scale", "scaleSub", 1.0f ) );

	AddParam( new ParamBool( "Do Add", "doAdd", true ) );
	AddParam( new ParamBool( "Do Sub", "doSub", true ) );

	// Hue, tolerance, feather, enabled (vec4 hue)
	AddParam( new ParamFloat( "Match Hue", ParameterType::FF_TYPE_HUE, "hue", 0.0f, ParamFloat::Grouping::VEC4 ) );
	AddParam( new ParamFloat( "Tol Hue", "", 0.05f ) );
	AddParam( new ParamFloat( "Soft Hue", "", 0.05f ) );
	AddParam( new ParamBool( "Use Hue", "", false ) );

	// Saturation range (min, max), feather, enabled (vec4 sat)
	AddParam( new ParamFloat( "Min Sat", ParameterType::FF_TYPE_SATURATION,  "sat", 0.5f, ParamFloat::Grouping::VEC4 ) );
	AddParam( new ParamFloat( "Max Sat", ParameterType::FF_TYPE_SATURATION, "", 1.0f ) );
	AddParam( new ParamFloat( "Soft Sat", "", 0.1f ) );
	AddParam( new ParamBool( "Use Sat", "", false ) );

	// Brightness range (min, max), feather, enabled (vec4 hue)
	AddParam( new ParamFloat( "Min Bri", ParameterType::FF_TYPE_BRIGHTNESS, "bri", 0.5f, ParamFloat::Grouping::VEC4 ) );
	AddParam( new ParamFloat( "Max Bri", ParameterType::FF_TYPE_BRIGHTNESS, "", 1.0f ) );
	AddParam( new ParamFloat( "Soft Bri", "", 0.1f ) );
	AddParam( new ParamBool( "Use Bri", "", false ) );

	AddParam( new ParamBool( "Show Mask", "showMask", false ) );
	AddParam( new ParamBool( "Cutout Mask", "cutoutMask", false ) );

	// Other tests

	//AddHSBAColorParams( magic, "Magic", "magic", { 0, 0, 0, 0 } );
	//magic.hue->doAutoConversion = false;

	//AddParam( new ParamFloat( "Test X pos", ParameterType::FF_TYPE_XPOS, "testX" ) );
	//AddParam( new ParamFloat( "Test Y pos", ParameterType::FF_TYPE_YPOS, "testY" ) );

	//AddParam( new ParamFloat( "TestAlpha", ParameterType::FF_TYPE_ALPHA ) );

	//AddParam( new ParamEvent( "TestEvent" ) );
	//AddAudioParam( new ParamFFT( "TestFFT", 8 ) );

	//testOption = new ParamOption( "TestOption", 2U );
	//testOption->AddOption( { "Opt 1", 11.1f } );
	//testOption->AddOption( { "Opt 2", 22.2f } );
	//testOption->AddOption( { "Opt 3", 33.3f } );
	//testOption->AddOption( { "Opt 4", 44.4f } );
	//testOption->AddOption( { "Opt 5", 55.5f } );
	//AddParam( testOption );

	//AddParam( new ParamRange( "TestRange", "temperature", 25.0f, { 0.0f, 100.0f } ) );
	//AddParam( new ParamInteger( "TestInteger", "xscale", 0, { -1920.0f, 1920.0f } ) );
	//AddParam( new ParamText( "TestText", "Lorem ipsum blah blah amet" ) );
	//AddParam( new ParamTrigger( "TestTrigger" ) );

	SetFragmentShader( _fragmentShaderCode );

	SetTimeSupported( true );
}

SubAdd::~SubAdd()
{
}

void SubAdd::Update()
{
	// Try out accessing parameter values

	float r0 = adjustment.red->GetFloatValue();
	float g0 = adjustment.green->GetFloatValue();
	float b0 = adjustment.blue->GetFloatValue();

	float average = ( r0 + g0 + b0 ) / 3.0f;

	RGBAColor rgb0 = adjustment.GetRGBColor();
	float red0     = rgb0.r;

	std::string uniformName1 = colAdd.hue->GetUniformName();

	HSBAColor color1 = colAdd.GetHSBAColor();
	HSBAColor color2 = colSub.GetHSBAColor();

	float hue1 = color1.hue;

	RGBAColor rgba2;
	colSub.GetHSBAColorAsRGBA( rgba2 );

	//HSBAColor hsbaMagic = magic.GetHSBAColor();
	//RGBAColor rgbaMagic = magic.GetHSBAColorAsRGBA();


	//FFUInt32 selectedOption   = testOption->GetIndex();
	//std::string selectedName  = testOption->GetOptionName( selectedOption );

	//float selectedValue;
	//FFResult res = testOption->GetOptionValue( selectedOption, selectedValue );

	//float selectedValueDirect = testOption->GetFloatValue();
}

}// namespace effect
}// namespace ffgl
}// namespace spxl
